import { h, defineComponent, ref, computed, watch, Transition, onBeforeUnmount, getCurrentInstance } from 'vue'

import useQuasar from '../../composables/use-quasar.js'
import useAnchor, { useAnchorProps } from '../../composables/use-anchor.js'
import useScrollTarget from '../../composables/use-scroll-target.js'
import useModelToggle, { useModelToggleProps, useModelToggleEmits } from '../../composables/use-model-toggle.js'
import useDark, { useDarkProps } from '../../composables/use-dark.js'
import usePortal from '../../composables/use-portal.js'
import useTransition, { useTransitionProps } from '../../composables/use-transition.js'
import useEmitListeners from '../../composables/use-emit-listeners.js'
import useTick from '../../composables/use-tick.js'
import useTimeout from '../../composables/use-timeout.js'

import { closePortalMenus } from '../../utils/portal.js'
import { getScrollTarget } from '../../utils/scroll.js'
import { position, stopAndPrevent, addEvt } from '../../utils/event.js'
import { hSlot } from '../../utils/composition-render.js'
import { addEscapeKey, removeEscapeKey } from '../../utils/escape-key.js'
import { addFocusout, removeFocusout } from '../../utils/focusout.js'
import { childHasFocus } from '../../utils/dom.js'
import { addClickOutside, removeClickOutside } from '../../utils/click-outside.js'

import {
  validatePosition, validateOffset, setPosition, parsePosition
} from '../../utils/position-engine.js'

export default defineComponent({
  name: 'QMenu',

  inheritAttrs: false,

  props: {
    ...useAnchorProps,
    ...useModelToggleProps,
    ...useDarkProps,
    ...useTransitionProps,

    persistent: Boolean,
    autoClose: Boolean,
    separateClosePopup: Boolean,

    noRouteDismiss: Boolean,
    noRefocus: Boolean,
    noFocus: Boolean,

    fit: Boolean,
    cover: Boolean,

    square: Boolean,

    anchor: {
      type: String,
      validator: validatePosition
    },
    self: {
      type: String,
      validator: validatePosition
    },
    offset: {
      type: Array,
      validator: validateOffset
    },

    scrollTarget: {
      default: void 0
    },

    touchPosition: Boolean,

    maxHeight: {
      type: String,
      default: null
    },
    maxWidth: {
      type: String,
      default: null
    }
  },

  emits: [
    ...useModelToggleEmits,
    'click', 'escape-key'
  ],

  setup (props, { slots, emit, attrs }) {
    let refocusTarget, absoluteOffset, unwatchPosition, avoidAutoClose

    const vm = getCurrentInstance()
    const innerRef = ref(null)
    const showing = ref(false)

    const hideOnRouteChange = computed(() =>
      props.persistent !== true &&
      props.noRouteDismiss !== true
    )

    const $q = useQuasar()
    const { isDark } = useDark(props, $q)
    const { registerTick, removeTick, prepareTick } = useTick()
    const { registerTimeout, removeTimeout } = useTimeout()
    const { transition } = useTransition(props, showing)
    const { localScrollTarget, changeScrollEvent, unconfigureScrollTarget } = useScrollTarget(props, configureScrollTarget)

    const { anchorEl, canShow } = useAnchor(props, {
      emit, vm, showing, $q
    })

    const { emitListeners } = useEmitListeners(vm)
    const { show, hide, toggle } = useModelToggle(props, {
      emit,
      showing, canShow, handleShow, handleHide,
      emitListeners,
      vm,
      hideOnRouteChange,
      processOnMount: true
    })

    const { showPortal, hidePortal, renderPortal } = usePortal(vm, renderPortalContent)

    const clickOutsideProps = {
      anchorEl,
      innerRef,
      getEl: () => vm.proxy.$el,
      onClickOutside (e) {
        if (props.persistent !== true && showing.value === true) {
          const targetClassList = e.target.classList

          hide(e)

          if (
            // always prevent touch event
            e.type === 'touchstart' ||
            // prevent click if it's on a dialog backdrop
            targetClassList.contains('q-dialog__backdrop')
          ) {
            stopAndPrevent(e)
          }

          return true
        }
      }
    }

    const anchorOrigin = computed(() =>
      parsePosition(
        props.anchor || (
          props.cover === true ? 'center middle' : 'bottom start'
        ),
        $q.lang.rtl
      )
    )

    const selfOrigin = computed(() =>
      props.cover === true
        ? anchorOrigin.value
        : parsePosition(props.self || 'top start', $q.lang.rtl)
    )

    const menuClass = computed(() =>
      (props.square === true ? ' q-menu--square' : '') +
      (isDark.value === true ? ' q-menu--dark q-dark' : '')
    )

    const onEvents = computed(() =>
      props.autoClose === true
        ? { onClick: onAutoClose }
        : {}
    )

    const handlesFocus = computed(() =>
      showing.value === true && props.persistent !== true
    )

    watch(handlesFocus, val => {
      if (val === true) {
        addEscapeKey(onEscapeKey)
        addClickOutside(clickOutsideProps)
      }
      else {
        removeEscapeKey(onEscapeKey)
        removeClickOutside(clickOutsideProps)
      }
    })

    function focus () {
      let node = innerRef.value

      if (node && node.contains(document.activeElement) !== true) {
        node = node.querySelector('[autofocus], [data-autofocus]') || node
        node.focus()
      }
    }

    function handleShow (evt) {
      removeTick()
      removeTimeout()

      refocusTarget = props.noRefocus === false
        ? document.activeElement
        : void 0

      addFocusout(onFocusout)

      showPortal()
      configureScrollTarget()

      absoluteOffset = void 0

      if (evt !== void 0 && (props.touchPosition || props.contextMenu)) {
        const pos = position(evt)

        if (pos.left !== void 0) {
          const { top, left } = anchorEl.value.getBoundingClientRect()
          absoluteOffset = { left: pos.left - left, top: pos.top - top }
        }
      }

      if (unwatchPosition === void 0) {
        unwatchPosition = watch(
          () => $q.screen.width + '|' + $q.screen.height + '|' + props.self + '|' + props.anchor + '|' + $q.lang.rtl,
          updatePosition
        )
      }

      if (props.noFocus !== true) {
        document.activeElement.blur()
      }

      registerTick(() => {
        updatePosition()
        props.noFocus !== true && focus()
      })
      prepareTick()

      registerTimeout(() => {
        // required in order to avoid the "double-tap needed" issue
        if ($q.platform.is.ios === true) {
          // if auto-close, then this click should
          // not close the menu
          avoidAutoClose = props.autoClose
          innerRef.value.click()
        }

        updatePosition()
        emit('show', evt)
      }, 300)
    }

    function handleHide (evt) {
      removeTick()
      removeTimeout()

      anchorCleanup(true)

      // check null for IE
      if (
        refocusTarget !== void 0 &&
        refocusTarget !== null &&
        (
          // menu was hidden from code or ESC plugin
          evt === void 0 ||
          // menu was not closed from a mouse or touch clickOutside
          evt.qClickOutside !== true
        )
      ) {
        refocusTarget.focus()
      }

      registerTimeout(() => {
        hidePortal()
        emit('hide', evt)
      }, 300)
    }

    function anchorCleanup (hiding) {
      absoluteOffset = void 0

      if (unwatchPosition !== void 0) {
        unwatchPosition()
        unwatchPosition = void 0
      }

      if (hiding === true || showing.value === true) {
        removeFocusout(onFocusout)
        unconfigureScrollTarget()
        if (props.persistent !== true) {
          removeClickOutside(clickOutsideProps)
          removeEscapeKey(onEscapeKey)
        }
      }
    }

    function configureScrollTarget () {
      if (anchorEl.value !== null || props.scrollTarget !== void 0) {
        localScrollTarget.value = getScrollTarget(anchorEl.value, props.scrollTarget)
        changeScrollEvent(localScrollTarget.value, updatePosition)
      }
    }

    function onAutoClose (e) {
      // if auto-close, then the ios double-tap fix which
      // issues a click should not close the menu
      if (avoidAutoClose !== true) {
        closePortalMenus(vm.proxy, e)
        emit('click', e)
      }
      else {
        avoidAutoClose = false
      }
    }

    function onFocusout (evt) {
      // the focus is not in a vue child component
      if (
        handlesFocus.value === true &&
        childHasFocus(innerRef.value, evt.target) !== true
      ) {
        focus()
      }
    }

    function onEscapeKey (evt) {
      emit('escape-key')
      hide(evt)
    }

    function updatePosition () {
      const el = innerRef.value

      if (anchorEl.value === null || !el) {
        return
      }

      setPosition({
        el,
        offset: props.offset,
        anchorEl: anchorEl.value,
        anchorOrigin: anchorOrigin.value,
        selfOrigin: selfOrigin.value,
        absoluteOffset,
        fit: props.fit,
        cover: props.cover,
        maxHeight: props.maxHeight,
        maxWidth: props.maxWidth
      })
    }

    function renderPortalContent () {
      return h(
        Transition,
        { name: transition.value, appear: true },
        () => showing.value === true
          ? h('div', {
              ...attrs, // TODO vue3 - verify reactivity
              ref: innerRef,
              tabindex: -1,
              class: [
                'q-menu q-position-engine scroll' + menuClass.value,
                attrs.class
              ],
              ...onEvents.value
            }, hSlot(slots.default))
          : null
      )
    }

    onBeforeUnmount(anchorCleanup)

    // expose public methods
    Object.assign(vm.proxy, {
      focus, show, hide, toggle, updatePosition,

      // expose needed stuff for portal utils
      quasarPortalInnerRef: innerRef
    })

    return renderPortal
  }
})
